/*
* This file is part of fwmon
* Copyright (c) 2001,2002 Gianni Tedesco
*
*    fwmon is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    fwmon is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with fwmon; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static const char rcsid[]=
	"@(#) $Id: sniffer.c,v 1.6 2002/02/04 08:35:03 gianni Exp $";

#include "g_globals.h"
#include "main.h"

#define __USE_XOPEN
#include <sys/poll.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <asm/system.h>
#include <linux/sockios.h>

#include <features.h>
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
# include <netpacket/packet.h>
# include <net/ethernet.h>
#else
# include <asm/types.h>
# include <linux/if_packet.h>
# include <linux/if_ether.h>
#endif

int frames;
int sniff_fd=-1;
char *map=NULL;
int mapsiz;
struct iovec *ring=NULL;

/* cache 32 interface names */
#define IFCACHE_SIZ 32
char *ifcache[IFCACHE_SIZ];

int sniffer_open(void)
{
	struct tpacket_req req;
	struct sockaddr_ll addr;
	int i;

	memset(ifcache, 0, IFCACHE_SIZ*sizeof(char *));	
	
	if ( (sniff_fd=socket(PF_PACKET, SOCK_DGRAM, 0))<0 ) {
		perror("socket()");
		return 0;
	}

	/* Setup PACKET_RX_RING */
	req.tp_block_size=4096;
	req.tp_frame_size=1024;
	req.tp_block_nr=64;
	req.tp_frame_nr=4*64;
	mapsiz=req.tp_block_size*req.tp_block_nr;
	frames=req.tp_frame_nr;
	if ( (setsockopt(sniff_fd, SOL_PACKET, PACKET_RX_RING,
		(char *)&req, sizeof(req))) != 0 ) {
		perror("setsockpt(PACKET_RX_RING)");
		return 0;
	}

	/* Do the memory map */
	if ( (map=mmap(NULL, mapsiz, PROT_READ|PROT_WRITE|PROT_EXEC,
		MAP_SHARED, sniff_fd, 0))==MAP_FAILED ) {
		perror("mmap() packet socket");
		close(sniff_fd);
		return 0;
	}

	/* Allocate the iovec ring */
	if ( !(ring=malloc(mapsiz)) ) {
		perror("allocating ring buffer");
		munmap(map, mapsiz);
		close(sniff_fd);
		return 0;
	}
	
	/* Initialise to our ring buffer */
	for(i=0; i<frames; i++) {
		ring[i].iov_base=(void *)((long)map)+(i*req.tp_frame_size);
		ring[i].iov_len=req.tp_frame_size;
	}

	/* Do the obligatory bind */
	memset(&addr, 0, sizeof(addr));
	addr.sll_family=AF_PACKET;
	addr.sll_protocol=htons(0x03);
	if ( bind(sniff_fd, (struct sockaddr *)&addr, sizeof(addr)) ) {
		perror("bind()");
		munmap(map, mapsiz);
		close(sniff_fd);
		return 0;
	}

	return 1;
}

const char *ifname(int ifindex)
{
	static struct ifreq req;

	/* Check the cache first */
	if ( ifindex < IFCACHE_SIZ && ifcache[ifindex]) {
		return ifcache[ifindex];
	}
	
	req.ifr_ifindex=ifindex;

	if ( ioctl(sniff_fd, SIOCGIFNAME, &req) ) {
		sprintf(req.ifr_name, "if%i", ifindex);
	}

	if ( ifindex < IFCACHE_SIZ ) {
		ifcache[ifindex]=strdup(req.ifr_name);
	}

	return req.ifr_name;
}

void sniffer_mainloop(void)
{
	int i;
	struct pollfd pfd;
	struct tpacket_hdr *h;
	struct sockaddr_ll *sll;
	struct netlink_t hdr;
	struct packet_t packet;

	packet.hdr=&hdr;
	hdr.mark=0;
		
	for(i=0;;) {
		while(*(unsigned long*)ring[i].iov_base) {
			h=ring[i].iov_base;
			sll=(void *)h + TPACKET_ALIGN(sizeof(*h));

			/* Only print IP packets */
			if ( sll->sll_protocol == 8 ) {
				hdr.len=h->tp_len;
				strcpy(hdr.iface, ifname(sll->sll_ifindex));	
				packet.ip=(void *)h + h->tp_mac;
				packet.ts.tv_sec=h->tp_sec;
				packet.ts.tv_usec=h->tp_usec;
				dump_packet(&packet);
			}
			
			h->tp_status=0;
			mb();
			i=(i==frames-1) ? 0 : ++i;
		}

		/* Sleep when not busy */
		pfd.fd=sniff_fd;
		pfd.events=POLLIN|POLLERR;
		pfd.revents=0;
		poll(&pfd, 1, -1);
	}
}

int sniffer_close(void)
{
	int i;

	for(i=0; i<IFCACHE_SIZ; i++) {
		if ( ifcache[i] ) free(ifcache[i]);
	}
	
	if ( map ) {
		munmap(map, mapsiz);
		map=NULL;
	}
	
	if ( sniff_fd>=0 ) {
		close(sniff_fd);
		sniff_fd=-1;
	}
	return 1;
}
